--*******************************************************************************************************************************
--
-- Description: This module performs all PCI Express write/read operations to registers not associated with a specific channel.
--				The AXI_WIDTH constant determines whether slave_dataout is 64-bit or 128-bit aligned.
--					
-- Rev History: 09-04-2013:	Inital release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.model_specific_package.all;
use work.sigfpga_package.all;

entity bar2_common is
port (                                                                                                                              
    -- Reset and clock.                                                                                      
	rst    				: in std_logic;									-- Active high asynchronous reset.
	clk		        	: in std_logic;									-- Bus clock.
	
    -- PCI Express write transaction interface.
	slv_addr 			: in std_logic_vector(15 downto 2);				-- Refer to the PLDA EZDMA2 reference manual.
    slv_dataout     	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	slv_bytevalid 		: in std_logic_vector((AXI_WIDTH/8)-1 downto 0);-- Refer to the PLDA EZDMA2 reference manual.
	bar2_write 			: in std_logic;									-- Logic one indicates that the current PCI write request is directed at BAR2.
                                                                                                                              
    -- PCI Express read transaction interface.
	bar2_slave_read			: in std_logic;								-- Logic one indicates that the current PCI read request is directed at BAR2.
	bar2_rddata_common	: out std_logic_vector(AXI_WIDTH-1 downto 0);	-- Data return from read operation on BAR2 common register.
    dma_rdaddr      	: in std_logic_vector(15 downto 3);				-- Refer to the PLDA EZDMA2 reference manual.
	                                                                                                                       	       	                                                                                                                       	       
	-- Command, status and error signals.
	soft_trig			: out std_logic;								-- Software trigger requested over the PCIe bus.
	led_control 		: out std_logic_vector(1 downto 0);  			-- Logic one illuminates each of the two LEDs.
	address_error 		: out std_logic_vector(1 downto 0)				-- Logic one indicates an invalid address for this decoder.
);                                                                                                                     	
end bar2_common;                                                                                                      	   
                                                                                                                               	 
architecture structural of bar2_common is
                                                                                                                               	
	signal bytevalid		: std_logic_vector(1 downto 0);				-- Logic one indicates that all PCI byte valids are active.
	signal led_control_reg	: std_logic_vector(1 downto 0);				-- Mask bits to disable specific sources from illuminating the LEDs.
	signal example			: DUAL_64;									-- Register to demonstrate 64-bit read and write transactions.

begin
	
--***************************************************** WRITE TRANSACTIONS *****************************************************

   process (clk,rst)
    begin
        if rst='1' then
            bytevalid <= (others => '0');
			address_error(0) <= '0';
			soft_trig <= '0';
        	led_control_reg <= (others => '0');
        	example(0) <= (others => '0');
        	example(1) <= (others => '0');

        elsif rising_edge(clk) then

        	-- Create an active high enable for write operations to the upper 32 register bits.
        	if slv_addr(3) = '0' then
        		bytevalid(1) <= (slv_bytevalid(7) and slv_bytevalid(6) and slv_bytevalid(5) and slv_bytevalid(4)) or slv_addr(2);
        	else
        		bytevalid(1) <= (slv_bytevalid((AXI_WIDTH/8)-1) and slv_bytevalid((AXI_WIDTH/8)-2) and slv_bytevalid((AXI_WIDTH/8)-3) and slv_bytevalid((AXI_WIDTH/8)-4)) or slv_addr(2);
			end if;        		
        	-- Create an active high enable for write operations to the lower 32 register bits.
        	if slv_addr(3) = '0' then
        		bytevalid(0) <= slv_bytevalid(3) and slv_bytevalid(2) and slv_bytevalid(1) and slv_bytevalid(0) and not(slv_addr(2));
        	else
        		bytevalid(0) <= slv_bytevalid((AXI_WIDTH/8)-5) and slv_bytevalid((AXI_WIDTH/8)-6) and slv_bytevalid((AXI_WIDTH/8)-7) and slv_bytevalid((AXI_WIDTH/8)-8) and not(slv_addr(2));
        	end if;
        
            -- Target write decoder that loads slave data into the selected register for a target write completion.
            if bar2_write = '1' then
                case slv_addr(15 downto 3) & "000" is
                
                	-- x"000" is a read only register.
					
					-- x"010" is defined below for single period trigger.
                    when x"0010" =>
                    	null;
						
					-- This is an example of a 2-bit register used to control LED illumination.
                    when x"0020" =>
						if bytevalid(0) = '1' then led_control_reg <= slv_dataout(1 downto 0); end if;
						
					-- This is an example of a 64-bit register aligned with a 128-bit address.
                    when x"0030" =>
						if bytevalid(1) = '1' then example(0)(63 downto 32) <= slv_dataout(63 downto 32); end if;
						if bytevalid(0) = '1' then example(0)(31 downto 0) <= slv_dataout(31 downto 0); end if;
						
					-- This is an example of a 64-bit register aligned with either a 64-bit or 128-bit address.
                    when x"0038" =>
						if bytevalid(1) = '1' then example(1)(63 downto 32) <= slv_dataout((AXI_WIDTH-64)+63 downto (AXI_WIDTH-64)+32); end if;
						if bytevalid(0) = '1' then example(1)(31 downto 0) <= slv_dataout((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64)); end if;
						
					-- Return an error if the address is not valid.
                    when others  =>
						address_error(0) <= '1';
						
                end case;    
			else
				address_error(0) <= '0';
            end if; 
            
            -- Generate a single period software trigger when there is a write to address x"010".
			if bar2_write = '1' and bytevalid(0) = '1' and (slv_addr(15 downto 3) & "000" = x"0010") then
				soft_trig <= '1';
			else
				soft_trig <= '0';
			end if;
			
        end if; 
    end process;
    
--***************************************************** READ TRANSACTIONS ******************************************************

    process (clk,rst)
    begin
        if rst = '1' then
            bar2_rddata_common <= (others => '0');
			address_error(1) <= '0';
        elsif rising_edge(clk) then
  
            -- Target read decoder that loads selected register contents onto the DMA bus for a target read completion.
            if bar2_slave_read = '1' then
                case dma_rdaddr(15 downto 3) & "000" is
                
					-- Model specific feature identification.
					when x"0000" =>
						bar2_rddata_common(63 downto 52) <= (others => '0');
						bar2_rddata_common(51 downto 48) <= std_logic_vector(to_unsigned(TX_CHANNEL_COUNT,4));
						bar2_rddata_common(47 downto 44) <= std_logic_vector(to_unsigned(RX_CHANNEL_COUNT,4));
						bar2_rddata_common(43 downto 32) <= MODEL;
						bar2_rddata_common(31 downto 0) <= REVISION;
						
                	--  x"010" is a write only register.
						
					-- This is an example of a 2-bit register used to control LED illumination.
                    when x"0020" =>
						bar2_rddata_common(63 downto 2) <= (others => '0');
						bar2_rddata_common(1 downto 0) <= led_control_reg;
                    
					-- This is an example of a 64-bit register aligned with a 128-bit address.
					when x"0030" =>
						bar2_rddata_common(63 downto 32) <= example(0)(63 downto 32);
						bar2_rddata_common(31 downto 0) <= example(0)(31 downto 0);
											
					-- This is an example of a 64-bit register aligned with either a 64-bit or 128-bit address.
					when x"0038" =>
						bar2_rddata_common((AXI_WIDTH-64)+63 downto (AXI_WIDTH-64)+32) <= example(1)(63 downto 32);
						bar2_rddata_common((AXI_WIDTH-64)+31 downto (AXI_WIDTH-64)) <= example(1)(31 downto 0);
										
					-- Return an error if the address is not valid.
                    when others  =>                                                                                                 
						bar2_rddata_common(31 downto 0) <= x"badaddb2";
						bar2_rddata_common <= (others => '0');
						address_error(1) <= '1';
						
	            end case;
			else
				address_error(1) <= '0';
            end if;
                        
        end if;
    end process;

--****************************************************** PORT ASSIGNMENTS ******************************************************

    led_control <= led_control_reg;
		
end structural;
