--******************************************************************************************************************************
--
-- Description: This controller performs serial bus transactions using different protocols.  The serial busses are used to
--				communicate with devices external to the FPGA.
--					
-- Rev History: 09-04-2013:	Initial release.
--
--******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;

entity serial_bus is
	port (
    -- Reset and clock.                                                                                      
	rst 			: in std_logic;						-- Active high asynchronous reset.
	clk 			: in std_logic;						-- Bus clock.

	-- Connections to command/status registers in the BAR0 module.
	serial_payload	: in std_logic_vector(7 downto 0);	-- Serial bus payload (instruction + data).
	serial_start	: in std_logic;						-- Logic one initiates a SPI bus transaction.
	serial_busy		: out std_logic; 					-- Logic one indicates that the process is active.
	serial_shift	: out std_logic; 					-- Logic one indicates that the register needs to perform a serial shift.
	serial_read		: out std_logic; 					-- Registered duplicate of sdo.
	payload_size	: in std_logic_vector(2 downto 0);	-- Size of the serial payload in bits.
	active_edge		: in std_logic;						-- Select active rising (1) or falling (0) serial clock.
	bus_select		: in std_logic_vector(2 downto 0);	-- Select the serial bus.
														-- 000: I2C
														-- 001: SPI
														-- 010: SPI chip select port expander (serial shift register)
														-- 011: Programmable synthesizer
														-- 100: GPIO/user translator port expander (serial shift register)
	
	--	The following signals connect to external devices through FPGA pins.
	i2c_scl 		: out std_logic;
	i2c_sda_in 		: in std_logic;
	i2c_sda_out 	: out std_logic;
	i2c_direction 	: out std_logic;
	spi_csb_cmd     : out std_logic_vector(2 downto 0);
    spi_clk			: out std_logic;
    spi_sdi			: out std_logic;
    spi_sdo			: in std_logic;
    synth_cmd 		: out std_logic_vector(2 downto 0);
    gpio_user_cmd 	: out std_logic_vector(2 downto 0)
	);
end serial_bus;

architecture structural of serial_bus is

	type serialstate is (idle, start, bit_shift, ack_nack, stop);	-- Type definitions for the SPI state machine.
	signal serial_state	: serialstate;  			-- Current serial machine state.
	signal bit_count	: unsigned(7 downto 0);		-- Counter used to track the number of bit shifts during a transaction.
	signal clk_divider	: unsigned(11 downto 0);	-- Counter used to divide the system clock to a lower frequency.
	signal data			: std_logic;				-- Serial data stream from the BAR0 register to the peripheral.
	signal word_count	: unsigned(1 downto 0);		-- Count of I2C words processed.
	signal i2c_cmd		: std_logic;				-- Start, stop and not acknowledge command bits for the I2C bus.
	signal read_write	: std_logic;				-- Read/write status of the I2C bus.
	signal serial_clk	: std_logic;				-- Serial clock for either rising or falling active edge devices.
	signal phase_90		: std_logic;				-- Logic one marks a quarter period of the serial clock.
	signal phase_180	: std_logic;				-- Logic one marks a half period of serial clock.
	signal phase_270	: std_logic;				-- Logic one marks three quarter period of the serial clock.
	signal phase_360	: std_logic;				-- Logic one marks the end of a serial clock period.
	
begin
  	
--*************************************************** SERIAL CLOCK GENERATOR ***************************************************

	-- Produce a serial output clock that does not exceed the original I2C maximum frequency (100 kHz).
	-- Note: This clock will be used with all peripherals even though they may support much higer frequencies.
	process (clk,rst)
	begin
		if rst = '1' then
			-- Initialize to one to simplify the state transition definitions.
			clk_divider <= (1 => '1', others => '0'); 
		   	
		elsif rising_edge(clk) then
		
			-- Create the serial clock by dividing the bus clock by 4096 (61 kHz @ clk = 250 MHz).
			if serial_state = idle then
				-- Initialize to one to simplify the state transition definitions.
				clk_divider <= (1 => '1', others => '0'); 
			else
				-- Increment the counter.
				clk_divider <= clk_divider + x"001";
			end if;
			
		end if;
	end process; 

	-- Create a serial clock that can be configured as rising (1) or falling (0) edge active.
	serial_clk <= clk_divider(11) when active_edge = '1' else not(clk_divider(11));
	-- Mark a quarter period of the serial clock.
	phase_90 <= '1' when clk_divider = x"400" else '0';
	-- Mark a half period of the serial clock (active edge).
	phase_180 <= '1' when clk_divider = x"800" else '0';
	-- Mark three quarter period of the serial clock.
	phase_270 <= '1' when clk_divider = x"C00" else '0';
	-- Mark the end of a serial clock period.
	phase_360 <= '1' when clk_divider = x"000" else '0';
		
--************************************************* STATE MACHINE TRANSITIONS **************************************************

	-- Create a finite state machine.
	process (clk,rst)
	begin
		if rst = '1' then
			serial_state <= idle;
		   	
		elsif rising_edge(clk) then
		
			-- Define each of the state machine transitions.
			case serial_state is
			
				-- Remain idle while waiting for the application logic to request a serial bus transaction.
				when idle =>
					-- Stay in this state for one complete serial clock period.
					if serial_start = '1' then
						serial_state <= start;
					end if;
		
				-- SPI: Pause to meet the minimum pulse width requirement imposed by some devices for chip select inactive.
				-- I2C: Issue the master start command.
				when start =>
					-- Wait for one complete serial clock period.
					if phase_360 = '1' then
						serial_state <= bit_shift;
					end if;
		
				-- Serial shift the requested number of bits (8/16/24/32/40/48/56/64).
				when bit_shift =>
					case payload_size is
						when "000" =>	-- 8 bits
							if phase_360 = '1' and bit_count = x"08" then
								-- The I2C protocol is built on sequential 8-bit transactions.		
								if bus_select = "000" then
									serial_state <= ack_nack;
								else
									serial_state <= stop;
								end if;
							end if;
						when "001" => 	-- 16 bits
							if phase_360 = '1' and bit_count = x"10" then
								serial_state <= stop;
							end if;
						when "010" => 	-- 24 bits
							if phase_360 = '1' and bit_count = x"18" then
								serial_state <= stop;
							end if;
						when "011" => 	-- 32 bits
							if phase_360 = '1' and bit_count = x"20" then
								serial_state <= stop;
							end if;
						when "100" => 	-- 40 bits
							if phase_360 = '1' and bit_count = x"28" then
								serial_state <= stop;
							end if;
						when "101" => 	-- 48 bits
							if phase_360 = '1' and bit_count = x"30" then
								serial_state <= stop;
							end if;
						when "110" => 	-- 56 bits
							if phase_360 = '1' and bit_count = x"38" then
								serial_state <= stop;
							end if;
						when "111" => 	-- 64 bits
							if phase_360 = '1' and bit_count = x"40" then
								serial_state <= stop;
							end if;
						when others => null;
					end case;
					
				-- I2C acknowledge and not acknowledge state following 8-bit payloads.
				when ack_nack =>
					-- Stay in this state for one complete serial clock period.
					if phase_360 = '1' then
						-- Stop when a write transaction completes in four words or a read transaction completes in three words.
						if word_count = "11" or (word_count = "10" and read_write = '1') then
							serial_state <= stop;
						-- There are additional words to process in this transaction.
						else
							serial_state <= bit_shift;
						end if;
					end if;
		
				-- SPI: Pause a quarter clock period to meet minimum hold time requirement imposed by some devices for chip select active.
				-- I2C: Issue the master stop command.
				when stop =>
					if (bus_select /= "000" and phase_90 = '1') or phase_360 = '1' then
						serial_state <= idle;
					end if;
		
				-- Define the default state to be idle.
				when others =>
					serial_state <= idle;
				
			end case;
		end if;
	end process; 

--*************************************************** STATE MACHINE EVENTS *****************************************************

	-- Define the events that cause a state machine transition.
	process (clk,rst)
	begin
		if rst = '1' then
			bit_count <= (others => '0'); 
			serial_shift <= '0';
			word_count <= (others => '0');
			read_write <= '0';
		   	
		elsif rising_edge(clk) then
		
			-- Transmit the serial payload data bits.
			if serial_state /= bit_shift then
				-- Disable serial data shifts.
				serial_shift <= '0';
				-- Reset the bit counter.
				bit_count <= (others => '0');
			-- Increment the bit counter after the active edge of serial clock but before the end of the serial clock period.
			elsif phase_270 = '1' then
				bit_count <= bit_count + x"01";
			-- Serial shift outgoing bus data by one bit following the inactive edge of serial clock.
			elsif phase_360 = '1' then
				serial_shift <= '1';
			-- Limit the serial shift to a single bit.
			else
				serial_shift <= '0';
			end if;
				
			-- Count the number of words processed in a single I2C transaction.
			if serial_state = idle then
				word_count <= (others => '0');
			-- Increment the counter upon exiting the ack_nack state.
			elsif serial_state = ack_nack and phase_360 = '1' then
				word_count <= word_count + "01";
			end if;

			-- Capture the last I2C payload bit of each word to determine if it is a read or write transaction.
			if serial_state = bit_shift and phase_180 = '1' then
				read_write <= data;
			end if;
			
		end if;
	end process; 
				
--************************************************ INTERNAL SIGNAL ASSIGNMENTS *************************************************

	-- Set the serial output bit required by the I2C start, stop and not acknowledge commands.
	process (clk,rst)
	begin
		if rst = '1' then
			i2c_cmd <= '0';
		elsif rising_edge(clk) then
			-- Create a falling edge start indicator and drive the output low during the first half of the stop state.
			if (serial_state = start and phase_270 = '1') or (serial_state = stop and phase_90 = '1') then
				i2c_cmd <= '0';
			-- Complete read transactions with a not acknowledge and create a rising edge stop indicator.
			elsif serial_state = idle or serial_state = ack_nack or (serial_state = stop and phase_270 = '1') then
				i2c_cmd <= '1';
			end if;
		end if;
	end process; 
				
	-- Select the data source bit position based on the requested payload size.
	-- Note: The I2C bus is a special case that consists of three 8-bit payloads in sequence.
	process (bus_select, payload_size, serial_payload)
	begin
		case payload_size is
			when "000" => if bus_select = "000" then data <= serial_payload(2); else data <= serial_payload(0); end if;
			when "001" => data <= serial_payload(1);
			when "010" => data <= serial_payload(2);
			when "011" => data <= serial_payload(3);
			when "100" => data <= serial_payload(4);
			when "101" => data <= serial_payload(5);
			when "110" => data <= serial_payload(6);
			when "111" => data <= serial_payload(7);
			when others => null;
		end case;	
	end process;
	
	-- Generate a busy flag throughout the complete bus transaction.
    serial_busy <= '0' when serial_state = idle else '1';
    
--************************************************** EXTERNAL PIN ASSIGNMENTS **************************************************

	-- Map the serial clock to the I2C clock port when it is selected.
    i2c_scl <= serial_clk when bus_select = "000" and (serial_state = start or serial_state = bit_shift or serial_state = ack_nack or serial_state = stop) else '0';
	-- Map the command bit or payload MSB to the I2C data output port.
	i2c_sda_out <= i2c_cmd when serial_state /= bit_shift else data;
	-- Set the I2C data pin direction to input (1) or output (0).
	i2c_direction <= '1' when bus_select /= "000" or serial_state = idle or serial_state = ack_nack or (serial_state = bit_shift and (word_count = "11" or (phase_360 = '1' and bit_count = x"08"))) else '0';

	-- Map the serial clock to the SPI clock port when it is selected.
    spi_clk <= serial_clk when serial_state = bit_shift and bus_select = "001" else not(active_edge);
	-- Map the payload MSB to the SPI data port.
	spi_sdi <= data;
	-- Map the serial clock to the SPI chip select command clock port when it is selected.
	spi_csb_cmd(0) <= serial_clk when bus_select = "010" and serial_state = bit_shift else '0';
	-- Map the payload MSB to the SPI chip select command data port.
	spi_csb_cmd(1) <= data;
	-- Enable the active low SPI chip select that was loaded into the port expander.
	spi_csb_cmd(2) <= '1' when bus_select = "001" and (serial_state = bit_shift or serial_state = stop) else '0';
	
	-- Map the serial clock to the synthesizer clock port when the synthesizer is selected.
    synth_cmd(0) <= serial_clk when bus_select = "011" and serial_state = bit_shift else not(active_edge);
	-- Map the payload MSB to the synthesizer data output port.
	synth_cmd(1) <= data;
	-- Set the synthesizer chip select active low.
	synth_cmd(2) <= '0' when bus_select = "011" and (serial_state = bit_shift or serial_state = stop) else '1';
	
	-- Map the serial clock to the GPIO/user translator command clock port when it is selected.
	gpio_user_cmd(0) <= serial_clk when bus_select = "100" and serial_state = bit_shift else '0';
	-- Map the payload MSB to the GPIO/user translator command data port.
	gpio_user_cmd(1) <= data;
	-- Load the GPIO/user translator port expander.
	gpio_user_cmd(2) <= '1' when bus_select = "100" and serial_state = stop else '0';
	
	-- Process incoming serial data from the peripheral chip.
	process (clk,rst)
	begin
		if rst = '1' then
			serial_read <= '0';
		elsif rising_edge(clk) then
			--  Register incoming serial data on the active serial clock edge.
			if phase_180 = '1' then
				--  Route the I2C or SPI port to the input register depending on selection.
				if bus_select = "000" then
					serial_read <= i2c_sda_in;
				else
					serial_read <= spi_sdo;
				end if;
			end if;
		end if;
	end process; 
				
end structural;
